# /// script
# requires-python = ">=3.13"
# dependencies = [
#     "libcst",
# ]
# ///

import libcst as cst
import libcst.matchers as m
from libcst.display import dump
from libcst import RemoveFromParent

from typing import Dict, Tuple, Self, Union
import wave as w
from os import getcwd


def get_audio_file_duration(sound_file_path: str) -> float:
    """
    Returns the length of the given .wav sound file path.
    """
    with w.open(sound_file_path, "r") as f:
        frames: int = f.getnframes()
        rate: int = f.getframerate()
        duration = frames / rate
        return round(duration, 2)


class GeminiTransformer(cst.CSTTransformer):
    """
    A class to add code to a Gemini generated code file.
    """

    def __init__(
        self: Self,
        sound_indicator_nodes: Dict[str, Tuple[str, float]],
    ) -> None:
        self.sound_indicator_nodes: Dict[str, Tuple[str, float]] = sound_indicator_nodes

    def leave_FunctionDef(
        self: Self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
    ) -> cst.FunctionDef:
        """
        This function adds `self.interactive_embed()` to the end of the construct function.
        """
        if original_node.name.value != "construct":
            return super().leave_FunctionDef(original_node, updated_node)

        interactive_code: cst.SimpleStatementLine = cst.parse_statement(
            "self.interactive_embed()"
        )
        new_body: cst.IndentedBlock = cst.IndentedBlock(
            body=[*updated_node.body.body, interactive_code]
        )
        return updated_node.with_changes(body=new_body)

    def leave_SimpleStatementLine(
        self: Self,
        original_node: cst.SimpleStatementLine,
        updated_node: cst.SimpleStatementLine,
    ) -> Union[cst.SimpleStatementLine, cst.FlattenSentinel]:
        """
        This function adds `self.add_sound(...)` after certain Manim function calls, such as `Create()` or `FadeOut()`.
        """
        # for child in original_node.children:
        #     # This for loop matches specific nodes to add `self.add_sound(...)` after lines containing
        #     # certain Manim function calls.
        #     for node, (
        #         sound_file_path,
        #         intensity,
        #     ) in self.sound_indicator_nodes.items():
        #         # First type of function call to match for.
        #         if m.matches(
        #             child,
        #             m.Expr(
        #                 value=m.Call(
        #                     func=m.Attribute(
        #                         value=m.Name(
        #                             value="self",
        #                         ),
        #                         attr=m.Name(
        #                             value="play",
        #                         ),
        #                     ),
        #                     args=[
        #                         m.ZeroOrMore(m.Arg()),
        #                         m.Arg(value=m.Call(func=m.Name(value=node))),
        #                         m.ZeroOrMore(m.Arg()),
        #                     ],
        #                 )
        #             ),
        #         ):
        #             run_time_arg = cst.Arg(
        #                 value=cst.Float(
        #                     value=str(get_audio_file_duration(sound_file_path))
        #                 ),
        #                 keyword=cst.Name(value="run_time"),
        #             )
        #             node_of_interest = updated_node.body[0]
        #             updated_args = node_of_interest.value.args + (run_time_arg,)

        #             updated_call = node_of_interest.value.with_changes(
        #                 args=updated_args
        #             )

        #             updated_body = [
        #                 node_of_interest.with_changes(value=updated_call)
        #             ] + list(updated_node.body[1:])

        #             updated_node = updated_node.with_changes(body=updated_body)

        #             sound_code: cst.SimpleStatementLine = cst.parse_statement(
        #                 f"self.add_sound('{sound_file_path}', {intensity})"
        #             )
        #             return cst.FlattenSentinel([sound_code, updated_node])

        return super().leave_SimpleStatementLine(original_node, updated_node)

    def leave_Arg(
        self: Self, original_node: cst.Arg, updated_node: cst.Arg
    ) -> Union[cst.Arg, cst.RemovalSentinel]:
        """
        Removes run_time=[value] from function calls for run_time=[sound_file_length] to be added later.
        """
        if m.matches(
            original_node,
            m.Arg(
                value=m.OneOf(m.Integer(), m.Float()),
                keyword=m.Name(value="run_time"),
            ),
        ):
            return RemoveFromParent()
        return updated_node


def add_interactivity(code: str, path: str = getcwd()) -> None:
    """
    Adds interactivity to the generated Gemini code.
    """
    code: cst.Module = cst.parse_module(code)

    # with open("cst_full_debug.txt", "w") as f:
    #     f.write(dump(code))

    sound_indicator_nodes: Dict[str, Tuple[str, int]] = {
        "Create": ("click.wav", 1),
        "Rotate": ("click.wav", 1),
        "FadeOut": ("click.wav", 1),
    }
    updated_cst: cst.Module = code.visit(GeminiTransformer(sound_indicator_nodes))

    print(f"Writing to {path}/generated_code.py...")
    with open(f"{path}/generated_code.py", "w") as f:
        f.write(updated_cst.code)

    print("Finished adding interactivity...")


# /// script
# requires-python = ">=3.13"
# dependencies = [
#     "httpx",
#     "ipython>=8.31.0",
#     "latex>=0.7.0",
#     "libcst>=1.7.0",
#     "manim>=0.19.0",
# ]
# ///

from asyncio import create_subprocess_exec
from os import getcwd
from httpx import AsyncClient, RequestError, Response
from subprocess import CalledProcessError
from typing import Dict
from .cst_parser import add_interactivity


async def run_manim_code(code: str, path: str = getcwd()) -> None:
    print("Adding interactivity...")
    add_interactivity(code, path)

    print("Running the scene...")
    try:
        proc = await create_subprocess_exec(
            "manim", "-pql", "generated_code.py", "--renderer=opengl"
        )
        await proc.wait()
    except FileNotFoundError:
        print("Could not find the generated code file.")
    except CalledProcessError as e:
        print(f"Error while running Manim: {e}")


async def generate_video(prompt: str, path: str = getcwd()) -> None:
    GEMINI_URL: str = "https://gemini-wrapper-nine.vercel.app/gemini"

    print("Getting response...")

    PROMPT: str = f"""Your sole purpose is to convert natural language into Manim code. 
You will be given some text and must write valid Manim code to the best of your abilities.
DON'T code bugs and SOLELY OUTPUT PYTHON CODE. Import ALL the necessary libraries.
Define ALL constants. After you generate your code, check to make sure that it can run.
Ensure all the generated manim code is compatible with manim 0.19.0. DO NOT USE
DEPRECATED CLASSES, such as "ParametricSurface." Ensure EVERY element in the scene is visually distinctive. 
Define EVERY function you use. Write text at the top to explain what you're doing.
REMEMBER, YOU MUST OUTPUT CODE THAT DOESN'T CAUSE BUGS. ASSUME YOUR CODE IS BUGGY, AND RECODE IT AGAIN.
The prompt: {prompt}"""

    async with AsyncClient() as client:
        try:
            response: Response = await client.post(GEMINI_URL, json={"prompt": PROMPT})
            response.raise_for_status()
        except RequestError as e:
            print(f"Error in getting the response: {e}")
            return

    if response.status_code != 200:
        print(f"Status Code Error: {response.status_code}")
        return

    json: Dict = response.json()

    if "error" in json:
        print(f"JSON Error: {json["error"]}")
        return

    code: str = json["output"]
    code = "\n".join(code.splitlines()[1:-1])

    print("Creating the interactive scene...")
    await run_manim_code(code, path)